import torch 
from tqdm import tqdm
from copy import deepcopy
import torchvision.transforms as transforms
from .dreamer_utils import Hook, default_func_mean

from .utils import (
    fft_to_rgb, 
    lucid_colorspace_to_rgb, 
    normalize, 
    show_rgb
)

from .transforms import random_resize
from .auto_image_param import auto_image_param

class dreamer():
    """wrapper over a pytorch model for visualization

        Args:
            model (torch.nn.Module): pytorch model 
            quiet (bool, optional): enable or disable progress bar. Defaults to True.
            device (str, optional): 'cpu' or 'cuda'. Defaults to 'cuda'.
    """
    def __init__(self, model, quiet = True, device = 'cuda'):
        
        self.model = model 
        self.model.eval()
        self.device = device
        self.model.to(self.device)
        self.default_func = default_func_mean
        self.transforms = None
        self.quiet = quiet

    def get_default_transforms(self, rotate, scale_max, scale_min, translate_x, translate_y):
        self.transforms= transforms.Compose([
            transforms.RandomAffine(degrees = rotate, translate= (translate_x, translate_y)),
            random_resize(max_size_factor = scale_max, min_size_factor = scale_min),
        ])

    def set_custom_transforms(self, transforms):
        self.transforms = transforms

    def render(self, layers, image_parameter = None,  width= 256, height = 256, iters = 120, lr = 9e-3, rotate_degrees = 15,  scale_max = 1.2,  scale_min = 0.5, translate_x = 0., translate_y = 0.,  custom_func = None, weight_decay = 0., grad_clip = 1.):
        """core function to visualize elements form within the pytorch model

        WARNING: width and height would be ignored if image_parameter is not None

        Args:
            layers (iterable): [model.layer1, model.layer2...]
            image_parameter: instance of torch_dreams.auto.auto_image_param
            width (int): width of image to be optimized 
            height (int): height of image to be optimized 
            iters (int): number of iterations, higher -> stronger visualization
            lr (float): learning rate
            rotate_degrees (int): max rotation in default transforms
            scale_max (float, optional): Max image size factor. Defaults to 1.1.
            scale_min (float, optional): Minimum image size factor. Defaults to 0.5.
            translate_x (float, optional): Maximum translation factor in x direction
            translate_y (float, optional): Maximum translation factor in y direction
            custom_func (function, optional): See docs for a better explanation. Defaults to None.
            weight_decay (float, optional): Weight decay for default optimizer. Helps prevent high frequency noise. Defaults to 0.
            grad_clip (float, optional): Maximum value of norm of gradient. Defaults to 1.

        Returns:
            image_parameter instance: To show image, use: plt.imshow(image_parameter.rgb)
        """
        if image_parameter is None:

            image_parameter = auto_image_param(height= height, width = width, device = self.device, standard_deviation = 0.01)
        else:
            image_parameter = deepcopy(image_parameter)

        if image_parameter.optimizer is None:
            image_parameter.get_optimizer(lr = lr, weight_decay = weight_decay)

        if self.transforms is None:
            self.get_default_transforms(rotate = rotate_degrees, scale_max = scale_max, scale_min= scale_min, translate_x = translate_x, translate_y = translate_y)

        hooks = []
        for layer in layers:
            hook = Hook(layer)
            hooks.append(hook)

        for i in tqdm(range(iters), disable= self.quiet):
            image_parameter.optimizer.zero_grad()

            img = fft_to_rgb(image_parameter.height, image_parameter.width, image_parameter.param, device= self.device)
            img = lucid_colorspace_to_rgb(img,device= self.device)
            img = torch.sigmoid(img)
            img = normalize(img, device= self.device)
            img = self.transforms(img)

            model_out = self.model(img)

            layer_outputs = []

            for hook in hooks:
                out = hook.output[0]
                layer_outputs.append(out)

            if custom_func is not None:
                loss = custom_func(layer_outputs)
            else:
                loss = self.default_func(layer_outputs)
            loss.backward()
            image_parameter.clip_grads(grad_clip= 1)
            image_parameter.optimizer.step()
        

        for hook in hooks:
            hook.close()

        return image_parameter
    
    def get_snapshot(self, layers, input_tensor):

        with torch.no_grad():
            input_tensor = input_tensor.to(self.device)

            hooks = []
            for layer in layers:
                hook = Hook(layer)
                hooks.append(hook)


            model_out = self.model(input_tensor.float())

            layer_outputs = []

            for hook in hooks:
                out = hook.output[0]
                layer_outputs.append(out)

        return layer_outputs


import torch.nn as nn
import torch
import numpy as np

from .utils import init_image_param, image_buf_to_rgb
import torchvision.transforms as transforms

class auto_image_param():
    def __init__(self, height, width, device, standard_deviation):
        self.height = height
        self.width = width
        self.param = init_image_param(height = height, width = width, sd = standard_deviation, device = device)
        self.param.requires_grad_()
        self.optimizer = None

    def fetch_optimizer(self, params_list, optimizer = None, lr = 1e-3, weight_decay = 0.):
        if optimizer is not None:
            optimizer = optimizer(params_list, lr = lr, weight_decay = weight_decay)
        else:
            optimizer = torch.optim.AdamW(params_list, lr=lr, weight_decay=weight_decay)
        return optimizer

    def get_optimizer(self, lr, weight_decay):

        self.optimizer = self.fetch_optimizer(
                                    params_list = [self.param],
                                    lr = lr,
                                    weight_decay= weight_decay
                                    )
    
    def clip_grads(self, grad_clip = 1.):
        torch.nn.utils.clip_grad_norm_(self.param,grad_clip)

    def to_hwc_tensor(self, device = 'cpu'):
        rgb = image_buf_to_rgb(h = self.height, w = self.width, img_buf = self.param , device= device).permute(1,2,0).to(device = device, dtype = torch.float32)
        return rgb

    def to_chw_tensor(self, device = 'cpu'):
        t = image_buf_to_rgb(h = self.height, w = self.width, img_buf = self.param , device= device)
        return t

    def __array__(self):
        return self.to_hwc_tensor().numpy()

    def save(self, filename):
        """Save an image_param as an image. Uses PIL to save the image

        usage:
            
            image_param.save(filename = 'my_image.jpg')
            

        Args:
            filename (str): image.jpg
        """
        ten = self.to_chw_tensor()
        im = transforms.ToPILImage()(ten)
        im.save(filename)


import numpy as np
import torch
from torch import tensor
from torchvision import transforms

def init_image_param(height , width, sd=0.01, device = 'cuda'):
    """Initializes an image parameter 

    Args:
        height (int): height of image
        width (int): width of image
        sd (float, optional): Standard deviation of pixel values. Defaults to 0.01.
        device (str): 'cpu' or 'cuda'

    Returns:
        torch.tensor: image param to backpropagate on
    """
    img_buf = np.random.normal(size=(1, 3, height, width//2 + 1, 2), scale=sd).astype(np.float32)
    spectrum_t = tensor(img_buf).float().to(device)
    return spectrum_t

def get_fft_scale(h, w, decay_power=.75, device = 'cuda'):
    d=.5**.5 # set center frequency scale to 1
    fy = np.fft.fftfreq(h,d=d)[:,None]
    if w % 2 == 1:
        fx = np.fft.fftfreq(w,d=d)[: w // 2 + 2]
    else:
        fx = np.fft.fftfreq(w,d=d)[: w // 2 + 1]        
    freqs = (fx*fx + fy*fy) ** decay_power
    scale = 1.0 / np.maximum(freqs, 1.0 / (max(w, h)*d))
    scale = tensor(scale).float()[None,None,...,None].to(device)
    return scale

def fft_to_rgb(height, width, image_parameter, device = 'cuda'):
    """convert image param to NCHW 

    WARNING: torch v1.7.0 works differently from torch v1.8.0 on fft. 
    Hence you might find some weird workarounds in this function.

    Latest docs: https://pytorch.org/docs/stable/fft.html

    Also refer:
        https://github.com/pytorch/pytorch/issues/49637

    Args:
        height (int): height of image
        width (int): width of image 
        image_parameter (auto_image_param): instance of class auto_image_param()

    Returns:
        torch.tensor: NCHW tensor

    size log:
        before: 
            torch.Size([1, 3, height, width//2, 2]) 
            OR 
            torch.Size([1, 3, height, width+1//2, 2])

        after: 
            torch.Size([1, 3, height, width])

    """
    scale = get_fft_scale(height, width, device = device)

    t = scale * image_parameter.to(device)

    if torch.__version__[:3] == "1.7":
        t = torch.irfft(t, 2, normalized=True, signal_sizes=(height,width))
    elif  torch.__version__[:3] == '1.8':
        """
        hacky workaround to fix issues for the new torch.fft on torch 1.8.0

        """
        t = torch.complex(t[..., 0], t[..., 1])
        t = torch.fft.irfftn(t, s = (3, height, width), dim = (1,2,3), norm = 'ortho')

    return t

def color_correlation_normalized():
    color_correlation_svd_sqrt = np.asarray([[0.26, 0.09, 0.02],
                                             [0.27, 0.00, -0.05],
                                             [0.27, -0.09, 0.03]]).astype(np.float32)
    max_norm_svd_sqrt = np.max(np.linalg.norm(color_correlation_svd_sqrt, axis=0))
    color_correlation_normalized = tensor(color_correlation_svd_sqrt / max_norm_svd_sqrt)
    return color_correlation_normalized

def lucid_colorspace_to_rgb(t,device = 'cuda'):

    t_flat = t.permute(0,2,3,1)
    # t_flat = torch.matmul(t_flat, color_correlation_normalized().T)
    t_flat = torch.matmul(t_flat , color_correlation_normalized().T.to(device))
    t = t_flat.permute(0,3,1,2)
    return t

def rgb_to_lucid_colorspace(t, device = 'cuda'):
    t_flat = t.permute(0,2,3,1)
    inverse = torch.inverse(color_correlation_normalized().T.to(device))
    t_flat = torch.matmul(t_flat, inverse)
    t = t_flat.permute(0,3,1,2)
    return t

def imagenet_mean_std(device = 'cuda'):
    return (tensor([0.485, 0.456, 0.406]).to(device), 
            tensor([0.229, 0.224, 0.225]).to(device))

def denormalize(x):
    mean, std = imagenet_mean_std()
    return x.float()*std[...,None,None] + mean[...,None,None]

def normalize(x, device = 'cuda'):
    mean, std = imagenet_mean_std(device = device)
    return (x-mean[...,None,None]) / std[...,None,None]

def image_buf_to_rgb(h, w, img_buf, device = 'cuda'):
    img = img_buf.detach()
    img = fft_to_rgb(h, w, img, device = device)
    img = lucid_colorspace_to_rgb(img, device=  device)
    img = torch.sigmoid(img)
    img = img[0]    
    return img
    
def show_rgb(img, label=None, ax=None, dpi=25, **kwargs):
    plt_show = True if ax == None else False
    if ax == None: _, ax = plt.subplots(figsize=(img.shape[2]/dpi,img.shape[1]/dpi))
    x = img.cpu().permute(1,2,0).numpy()
    ax.imshow(x)
    ax.axis('off')
    ax.set_title(label)
    if plt_show: plt.show()

def gpu_affine_grid(size):
    size = ((1,)+size)
    N, C, H, W = size
    grid = torch.FloatTensor(N, H, W, 2).cuda()
    linear_points = torch.linspace(-1, 1, W) if W > 1 else tensor([-1.])
    grid[:, :, :, 0] = torch.ger(torch.ones(H), linear_points).expand_as(grid[:, :, :, 0])
    linear_points = torch.linspace(-1, 1, H) if H > 1 else tensor([-1.])
    grid[:, :, :, 1] = torch.ger(linear_points, torch.ones(W)).expand_as(grid[:, :, :, 1])
    return vision.FlowField(size[2:], grid)

def tensor_stats(t, label=""):
    if len(label) > 0: label += " "
    return("%smean:%.2f std:%.2f max:%.2f min:%.2f" % (label, t.mean().item(),t.std().item(),t.max().item(),t.min().item()))


import torch
import numpy as np

def default_func_MSE(layer_outputs):
    losses = []
    for output in layer_outputs:

        loss_component = torch.nn.MSELoss(reduction='mean')(output, torch.zeros_like(output))
        losses.append(loss_component)

    loss = torch.mean(torch.stack(losses))
    return -loss

def default_func_norm(layer_outputs):
    losses = []
    for output in layer_outputs:
        losses.append(output.norm())
    loss = torch.mean(torch.stack(losses))
    return -loss

def default_func_mean(layer_outputs):
    losses = []
    for output in layer_outputs:
        losses.append(output.mean())
    loss = torch.mean(torch.stack(losses))
    return -loss

class Hook():
    def __init__(self, module, backward=False):
        if backward==False:
            self.hook = module.register_forward_hook(self.hook_fn)
        else:
            self.hook = module.register_backward_hook(self.hook_fn)
    def hook_fn(self, module, input, output):
        self.input = input
        self.output = output
    def close(self):
        self.hook.remove()

def get_gradients(net_in, net, layers, default_func, custom_func = None):   
        """
        Executes the forward pass through the model and returns the gradients from the selected layer. 

        input args{
            net_in:  the 3D tensor which is to be used in the forward pass <size = (C, H, W)>
            net:  pytorch model which is being used for the  deep-dream
            layer:  layer instance of net whose activations are to be maximized
        }

        returns{
            gradient of model weights 
        }
        """  
        net_in.requires_grad = True
        net.zero_grad()

        hooks = []
        for layer in layers:

            hook = Hook(layer)
            hooks.append(hook)

        net_out = net(net_in)

        layer_outputs = []

        for hook in hooks:

            out = hook.output[0]
            layer_outputs.append(out)

        if custom_func is not None:
            loss = custom_func(layer_outputs)
        else:
            loss = default_func(layer_outputs)


        loss.backward()
        return net_in.grad.data

def make_octave_sizes(original_size, num_octaves, octave_scale):
    
    sizes = []

    for n in range(-num_octaves, 1):
        octave_size = tuple( np.array(original_size) * octave_scale**n)
        new_size = (int(octave_size[1]), int(octave_size[0]))

        sizes.append(new_size)

    return sizes


import torchvision.transforms as transforms
import torch.nn as nn
import random
from .image_transforms import resize_4d_tensor_by_factor, resize_4d_tensor_by_size

class random_resize(nn.Module):
    def __init__(self, max_size_factor, min_size_factor):
        super().__init__()
        self.max_size_factor = max_size_factor
        self.min_size_factor = min_size_factor

    def forward(self, x):
    
        # size = random.randint(a = 300, b = 600)
        # resized= resize_4d_tensor_by_size(x = x, height = size, width = size)

        height_factor = random.uniform(a = self.min_size_factor , b = self.max_size_factor)
        width_factor = random.uniform(a = self.min_size_factor , b = self.max_size_factor)

        resized = resize_4d_tensor_by_factor(x = x, height_factor = height_factor, width_factor = width_factor)
        return resized


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt') as f:
    required = f.read().splitlines()

setuptools.setup(
    name="torch-dreams",
    version="2.0.2",
    author="Mayukh Deb", 
    author_email="mayukhmainak2000@gmail.com", 
    description= "Making neural networks more interpretable, for research and art",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Mayukhdeb/torch-dreams",
    packages=setuptools.find_packages(),
    install_requires= required,
    python_requires='>=3.6',   
    include_package_data=True,
    keywords=[
        "PyTorch",
        "machine learning",
        "neural networks",
        "convolutional neural networks",
        "feature visualization",
        "optimization",
        ],
    classifiers=[
        "Intended Audience :: Science/Research",
        "Intended Audience :: Education",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    test_suite='nose.collector',
    tests_require=['nose']   
)



from .dreamer import *
from .utils import *
from .models import *
from .model_bunch import *
from .tests import *

__version__ = "2.0.2"

__all__ = [
    "dreamer",
    "utils",
    "models",
    "model_bunch",
    "auto_image_param",
    "image_transforms",
    "transforms"
]
